% function [B,C,Q,R,Mu,LL]=fcfa(X,D,K,cyc,tol);
% 
% Fast Conditional Factor Analysis (FA with inputs)
%
% X - N x p data matrix
% D - number of inputs (rest assumed to be outputs; default 0)
% K - size of state space (default 2)
% cyc - maximum number of cycles of EM (default 100)
% tol - termination tolerance (prop change in likelihood) (default 0.0001)
%
% B - input matrix
% C - observation (output) matrix 
% Q - state noise covariance 
% R - observation noise covariance
% Mu - input/output mean
% LL - log likelihood curve
%
% Iterates until a proportional change < tol in the log likelihood 
% or cyc steps of EM
%
% Random Init !!

function [B,C,Q,R,Mu,LL]=fcfa(X,D,K,cyc,tol);

p=length(X(1,:));
N=length(X(:,1));
tiny=exp(-700);

if nargin<5   tol=0.0001; end;
if nargin<4   cyc=100; end;
if nargin<3   K=2; end;
if nargin<2   D=0; end;

Mu=mean(X);
X=X-ones(N,1)*Mu;

if (D==0)
  [L,Ph,LL]=ffa(X,K,cyc,tol);
  B=[];
  C=L;
  Q=eye(K);
  R=Ph;
else
  
  % divide into inputs and outputs 
  U=X(:,1:D); % N*D
  Y=X(:,D+1:p);
  p=p-D;
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % initialize with Linear Regression
  
  fprintf('\nInitializing Randomly...\n');

  Q=eye(K);
  B=randn(K,D);
  C=randn(p,K);
  R=ones(p,1);

  fprintf('\nInitialized.\n');
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  I=eye(K);
  Y=Y';             % pxN
  U=U';             % DxN
  lik=0;
  LL=[];
  const=(2*pi)^(-p/2);

  UU=U*U';
  YY=Y*Y';
  YU=Y*U';
  
  Pfin=zeros(K,K);   % Var(x|y)
  Pt=zeros(K,K);     % E(x x' | y)
  Kcur=zeros(K,p);
  invP=zeros(p,p);

  for cycle=1:cyc

    % E STEP
    
    invR=diag(1./(R+(R<tiny)*tiny));
    
    tmp1=diag(R)+C*Q*C';
    
    if (K<p)
      T1=invR*C;
      invP=invR-T1*Q*inv(I+C'*T1*Q)*T1';
    else
      invP=inv(tmp1); % diag(R)+C*Q*C'
    end;
    Kcur=Q*C'*invP; %Kxp
    Pfin=Q-Kcur*C*Q;
    
    % Calculate Likelihood 

    % keyboard;
    
    oldlik=lik;
    lik=0;
    MM=invP;
    % dM=sqrt(det(MM));
    ldM=-0.5*sum(log(svd(tmp1)));
    
    if (isreal(ldM))
      CB=C*B;
      temp=CB*YU';
      lik=N*log(const)+N*ldM-0.5*sum(diag(MM*(YY-temp'-temp+CB*UU*CB'))); 
    else
      fprintf(' problem '); 
    end;

    LL=[LL lik];

    fprintf('cycle %g lik %g',cycle,lik);

    if (cycle<=2)
      likbase=lik;
    elseif (lik<oldlik) 
      fprintf(' violation');
    elseif ((lik-likbase)<(1 + tol)*(oldlik-likbase)|~finite(lik)) 
      fprintf('\n');
      break;
    end;
    fprintf('\n');
    
    % M STEP
    % Re-estimate B,C,Q,R;

    KCB=Kcur*C*B;
    temp=B-KCB;
    temp2=Kcur*YU*temp';
    XX=temp*UU*temp'+temp2+temp2'+Kcur*YY*Kcur';
    
    Pt=Pfin+XX/N;

    YX=YU*(B-KCB)'+YY*Kcur';
    C=YX*inv(Pt)/N;
    R=diag(YY)/N-diag(C*YX')/N;

    XU=(B-KCB)*UU+Kcur*YU;
    B=XU*inv(UU);
    
    Q1=Pt-B*XU'/N;
    Q=diag(diag(Q1));

    if (det(Q)<0 | ~symm(Q)) 
      fprintf('Q problem\n');
    end;
    Q=(Q+Q')/2;
    % keyboard;
  end;
end;

